import json
import codecs
from pathlib import Path
import argparse
import uuid
import shutil
from urllib.parse import urlparse
import traceback
from lyrebird.log import get_logger


logger = get_logger()


NEW_INFO_FILENAME = '.lyrebird_prop'
OLD_INFO_FILENAME = 'conf.json'


class WorkDirNotExists(Exception):
    pass


class WorkDirIsNotDir(Exception):
    pass


def update(work_dir, output_dir):
    if not work_dir.exists():
        raise WorkDirNotExists(f'{work_dir} not exists')
    if not work_dir.is_dir():
        raise WorkDirIsNotDir(f'{work_dir} is not dir')

    output_dir.mkdir(parents=True, exist_ok=True)

    work_dir = work_dir.expanduser().absolute()
    output_dir = output_dir.expanduser().absolute()
    print(f'work dir is {work_dir} Output dir is {output_dir}')
    
    for group_dir_name in Path(work_dir).iterdir():
        try:
            update_group(work_dir/group_dir_name.name, output_dir/group_dir_name.name)
        except Exception as e:
            logger.error(f'!!! Load group fail. {e}. {group_dir_name} . [removed]')
            traceback.print_exc()
            shutil.rmtree(str(output_dir/group_dir_name.name))


def update_group(group_dir, output_group_dir):
    old_conf_file = group_dir/OLD_INFO_FILENAME
    if not old_conf_file.exists():
        logger.error(f'Dir {group_dir} is not data group. skip >>')
        return

    print(f'Working in group {group_dir}')
    with codecs.open(old_conf_file, 'r', 'utf-8') as f:
        old_conf = json.load(f)

    output_group_dir.mkdir(exist_ok=True)
    print(f'Copy to target {output_group_dir}')

    new_group_info = {
        "id": str(uuid.uuid4()),
        "name": group_dir.name,
        "parent": None
    }
    with codecs.open(output_group_dir/NEW_INFO_FILENAME, 'w', 'utf-8') as f:
        json.dump(new_group_info, f, ensure_ascii=False, indent=4)

    saved_filter = set()
    for data_filter in old_conf['filters']:
        data_name = data_filter.get('response')
        if not data_name:
            logger.error(f'Bad data: not set response data. {data_filter}')
            continue
        contents = data_filter.get('contents')
        if not contents:
            logger.error(f'Bad data: not set filter. {data_filter}')
            continue
        filter_key = '.*'.join(contents)
        if filter_key in saved_filter:
            logger.error(f'Bad data: duplicate filter. {data_filter} -[removed]')
            continue
        try:
            update_data(group_dir/data_name, output_group_dir/data_name, old_group_filter=contents)
            saved_filter.add(filter_key)
        except Exception as e:
            logger.error(f'!!! Load data error. {e} . {group_dir/data_name}')
            traceback.print_exc()


def update_data(data_dir, output_data_dir, old_group_filter=None):
    print(f'Working in data {data_dir}')

    output_data_dir.mkdir(exist_ok=True)

    req = data_dir/'request.json'
    req_data = data_dir/'request_data.json'
    req_data_bin = data_dir/'request_data.bin'

    resp = data_dir/'response.json'
    resp_data = data_dir/'response_data.json'
    resp_data_bin = data_dir/'response_data.bin'

    req_obj = {}
    with codecs.open(req, 'r', 'utf-8') as f:
        req_obj = json.load(f)
    with codecs.open(output_data_dir/'request', 'w', 'utf-8') as f:
        json.dump({
            'url': req_obj.get('url'),
            'method': req_obj.get('method'),
            'headers': req_obj.get('headers')
        }, f, ensure_ascii=False, indent=4)
    
    if req_data.exists():
        shutil.copy(req_data, output_data_dir/'request_data')
    elif req_data_bin.exists():
        shutil.copy(req_data_bin, output_data_dir/'request_data')
    
    resp_obj = {}
    with codecs.open(resp, 'r', 'utf-8') as f:
        resp_obj = json.load(f)
    with codecs.open(output_data_dir/'response', 'w', 'utf-8') as f:
        json.dump({
            "code": resp_obj.get('code'),
            "headers": resp_obj.get('headers')
        }, f, ensure_ascii=False, indent=4)
    
    if resp_data.exists():
        shutil.copy(resp_data, output_data_dir/'response_data')
    elif resp_data_bin.exists():
        shutil.copy(resp_data_bin, output_data_dir/'response_data')

    new_data_info = {
                'id': str(uuid.uuid4()),
                'name': "",
                'rule': {},
                'request_data_type': None,
                'response_data_type': None
            }

    url = req_obj.get('url')
    if url:
        parsed_url = urlparse(url)
        if parsed_url.path != '/' and parsed_url.path != '':
            new_data_info['name'] = parsed_url.path
        else:
            new_data_info['name'] = parsed_url.hostname

    regex = ''
    for filter_line in old_group_filter:
        regex += f'(?=.*{filter_line})'
    new_data_info['rule'] = {'request.url': regex}

    content_type = req_obj.get('headers', {}).get('Content-Type')
    if content_type:
        if 'json' in content_type:
            new_data_info['request_data_type'] = 'json'
        elif 'html' in content_type:
            new_data_info['request_data_type'] = 'html'
        elif 'xml' in content_type:
            new_data_info['request_data_type'] = 'xml'
        elif 'text' in content_type:
            new_data_info['request_data_type'] = 'text'
    
    content_type = resp_obj.get('headers', {}).get('Content-Type')
    if content_type:
        if 'json' in content_type:
            new_data_info['response_data_type'] = 'json'
        elif 'html' in content_type:
            new_data_info['response_data_type'] = 'html'
        elif 'xml' in content_type:
            new_data_info['response_data_type'] = 'xml'
        elif 'text' in content_type:
            new_data_info['response_data_type'] = 'text'
    
    with codecs.open(output_data_dir/NEW_INFO_FILENAME, 'w', 'utf-8') as f:
        json.dump(new_data_info, f, ensure_ascii=False, indent=4)


def check_data_dir(data_root_path):
    data_dir = Path(data_root_path)
    data_dir.mkdir(exist_ok=True)
    if (data_dir/NEW_INFO_FILENAME).exists():
        return
    # Move data dir for data format
    old_dir = data_dir.parent/(data_dir.name+'(old)')
    shutil.move(data_dir, old_dir)
    # Transform old data
    update(old_dir, data_dir)
    # Touch new data prop file
    (data_dir/NEW_INFO_FILENAME).touch()

def main():
    parser = argparse.ArgumentParser(prog='DataFormatUpdateHelper')
    parser.add_argument('-d', '--dir', dest='root_dir', help='Set target data root dir')
    parser.add_argument('-o', '--output', dest='output_dir', help='Set output root dir')

    args = parser.parse_args()
    if not args.root_dir:
        work_dir = Path('.')
    else:
        work_dir = Path(args.root_dir)
    if not args.output_dir:
        output_dir = Path('./output')
    else:
        output_dir = Path(args.output_dir)
    update(work_dir, output_dir)


if __name__ == '__main__':
    main()



from flask_restful import Resource
from lyrebird.mock import context
from flask import request, jsonify, abort
from lyrebird import application

class Event(Resource):

    def post(self):
        msg = request.json.get('eventInfo')

        for action in msg.get('actions'):
            if action.get('type') == 'carrier':
                
                carrier_message = action.get('box', {})
                issue_channel = carrier_message.get('channel')
                issue_message = carrier_message.get('message')
                issue_state = carrier_message.get('state')
                
                issue_message['sender'] = msg.get('sender')
                
                application.server['event'].publish(issue_channel, issue_message, state=issue_state)
        return context.make_ok_response()



from pathlib import Path
import json
from .group import Group
from .exceptions import DataRootDirNotExistsError, ActivateFailed
from .mock_router import MockRouter
from lyrebird.log import get_logger


logger = get_logger()


class DataManager:

    def __init__(self):
        self._root_path = None
        self.activated_group_id = None
        self.groups = {}
        self.router = MockRouter()

    @property
    def root(self):
        return self._root_path

    @root.setter
    def root(self, root_path):
        new_path = Path(root_path).expanduser()
        if not new_path.exists():
            new_path.mkdir(parents=True)
        self._root_path = new_path
        self.scan()

    def activate(self, group_id):
        if group_id in self.groups:
            self.activated_group_id = group_id
            group = self.groups.get(group_id)
            parent = self.groups.get(group.parent_id)
            self.router.switch_group(group, parent=parent)
        else:
            raise ActivateFailed(f'Group id not found {group_id}')

    def deactivate(self):
        self.activated_group_id = None
        self.router.switch_group(None)

    def create_group(self):
        group = Group.new_group(self.root)
        self.groups[group.id] = group
        return group

    def delete_group(self, group_id):
        group = self.groups.pop(group_id)
        if group:
            group.delete()

    def copy_croup(self):
        pass
    
    def scan(self):
        for sub_file in self.root.iterdir():
            if not sub_file.is_dir():
                continue
            try:
                group = Group.createify(sub_file)
                if group:
                    self.groups[group.id] = group
            except Exception as e:
                logger.error(f'Load group failed. {sub_file}\n{e}')    
    



import json
import datetime
import traceback
from queue import Queue
from lyrebird import application
from lyrebird.base_server import ThreadServer
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer, Text, DateTime, create_engine

"""
Database server

Worked as a backgrund thread
storage all event message into database
"""

Base = declarative_base()

class LyrebirdDatabaseServer(ThreadServer):
    def __init__(self):
        super().__init__()

        # init sqlite
        DB_FILE_NAME = 'lyrebird.db'
        ROOT_DIR = application.root_dir()
        SQLALCHEMY_DATABASE_URI = ROOT_DIR/DB_FILE_NAME
        sqlite_path = 'sqlite:///'+str(SQLALCHEMY_DATABASE_URI)
        # TODO: 'sqlite:///' is unfriendly to windows

        engine = create_engine(str(sqlite_path))
        Base.metadata.create_all(engine)

        DBSession = sessionmaker(bind=engine)
        self.session = DBSession()

        # init queue
        self.storage_queue = Queue()
        
        # subscribe all channel
        application.server['event'].subscribe('any', self.put_queue, event=True)

    def put_queue(self, event):
        message, channel = event.message, event.channel
        content = json.dumps(message)
        flow = FLow(channel=channel, content=content)
        self.storage_queue.put(flow)

    def run(self):
        while self.running:
            try:
                flow = self.storage_queue.get()
                self.session.add(FLow(channel=flow.channel, content=flow.content))
                self.session.commit()
            except Exception:
                traceback.print_exc()

    def stop(self):
        super().stop()


class FLow(Base):
    __tablename__ = 'flow'

    id = Column(Integer, primary_key=True, autoincrement=True)
    channel = Column(String(16))
    content = Column(Text)
    ts = Column(DateTime(timezone=True), default=datetime.datetime.now)



import codecs
import json
import os
import traceback
import genson
from types import FunctionType
from flask import Blueprint, request, abort

from ..handlers.handler_context import HandlerContext
from .. import plugin_manager
from .. import context
from lyrebird import log


logger = log.get_logger()


api_mock = Blueprint('mock', __name__, url_prefix='/mock')


@api_mock.route('/')
@api_mock.route('/<path:path>', methods=['GET', 'POST', 'HEAD', 'PUT', 'DELETE', 'OPTIONS'])
def index(path=''):
    # todo add request handlers
    resp = None
    req_context = HandlerContext(request, path)
    req_context.update_client_req_time()

    for handler_name in plugin_manager.inner_handler:
        handler = plugin_manager.inner_handler[handler_name]
        try:
            handler.handle(req_context)
            if req_context.response:
                resp = req_context.response
        except Exception:
            logger.error(traceback.format_exc())

    for plugin_name in plugin_manager.data_handler_plugins:
        try:
            plugin = plugin_manager.data_handler_plugins[plugin_name]
            plugin.handle(req_context)
            if hasattr(plugin, 'change_response'):
                if isinstance(plugin.change_response, bool) and plugin.change_response:
                    resp = req_context.response
                elif isinstance(plugin.change_response, FunctionType) and plugin.change_response():
                    resp = req_context.response
                else:
                    logger.error(f'Plugin {plugin_name} has attr change_response, but its not bool or function')
        except Exception:
            logger.error(f'plugin error {plugin_name}\n{traceback.format_exc()}')

    if not resp:
        resp = abort(404, 'Not handle this request')
    
    req_context.update_client_resp_time()
    
    context.emit('action', 'add flow log')

    return resp


def create_json_schema(response_json_file_path):
    json_obj = json.loads(codecs.open(response_json_file_path, 'r', 'utf-8').read())
    schema = genson.Schema()
    schema.add_object(json_obj)
    schema_file_path = os.path.join(os.path.dirname(response_json_file_path), 'schema.json')
    schema_file = codecs.open(schema_file_path, 'w', 'utf-8')
    schema_file.write(schema.to_json(ensure_ascii=False, indent=4))
    schema_file.close()







from pathlib import Path
import codecs
import json
from packaging import version
from urllib.parse import urlparse
import requests
import subprocess
import time
import shutil
import jinja2
import json
import os
from lyrebird import log as nlog


logger = nlog.get_logger()


config_template = {
  "version": "0.10.5",
  "proxy.filters": [],
  "proxy.port": 4272,
  "mock.port": 9090,
  "mock.data": "data",
  "mock.proxy_headers": {
    "scheme": "MKScheme",
    "host": "MKOriginHost",
    "port": "MKOriginPort"
  }
}


class ConfigManager():
    ROOT = Path('~/.lyrebird').expanduser()
    DEFAULT_FILENAME = 'conf.json'
    BASE_CONFIG = ROOT/DEFAULT_FILENAME

    def __init__(self, conf_path=None):
        self.update_base_config()
        self.root = self.ROOT
        self.config = None
        self.conf_file = None
        if conf_path:
            self.update_conf(conf_path)
        self.config = self.read_base_config()
        if conf_path:
            self.read()        

    def update_conf(self, path):
        input_path:Path = Path(path).expanduser().absolute()
        if input_path.is_dir():
            self.root = input_path
            self.conf_file = input_path / self.DEFAULT_FILENAME
        else:
            self.root = input_path.parent
            self.conf_file = input_path

        # load config or use default config
        if not self.conf_file.exists():
            raise ConfigException(f'{self.conf_file} not found')

    def read(self):
        template_env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(self.root)))
        template = template_env.get_template(self.conf_file.name)
        custom_config = json.loads(template.render(current_dir=str(self.root), download_dir=str(self.ROOT/'downloads')))
        self.config.update(custom_config)

    def save(self):
        with codecs.open(self.conf_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.config, ensure_ascii=False, indent=4))
    
    def update_base_config(self):
        if self.BASE_CONFIG.exists() and self.BASE_CONFIG.is_file():
            with codecs.open(self.BASE_CONFIG, 'r', 'utf-8') as f:
                base_conf = json.load(f)
                if version.parse(base_conf.get('version', '0.0.0')) < version.parse(config_template.get('version', '0.0.0')):
                    self.write_base_config()
        else:
            self.write_base_config()

    def write_base_config(self):
        self.ROOT.mkdir(parents=True, exist_ok=True)
        with codecs.open(self.BASE_CONFIG, 'w', 'utf-8') as f:
            f.write(json.dumps(config_template, indent=4, ensure_ascii=False))

    def read_base_config(self):
        template_env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(self.ROOT)))
        template = template_env.get_template('conf.json')
        base_config = json.loads(template.render(current_dir=str(self.root), download_dir=str(self.ROOT/'downloads')))
        return base_config


class ConfigException(Exception):
    pass


resource_template = {
    'uri': None,
    'config': None
}


class Rescource:
    cache_filename = 'resource.json'
    download_dirname = 'downloads'

    def __init__(self, conf_root_path='~/.lyrebird'):
        self.root = Path(conf_root_path).expanduser().absolute()
        self.cache_file = self.root / self.cache_filename
        self.download_dir = self.root / self.download_dirname
        self.cache = None
        # load cache
        self.load()

    def load(self):
        if self.cache_file.exists():
            with codecs.open(self.cache_file, 'r', 'utf-8') as f:
                try:
                    self.cache = json.load(f)
                except Exception:
                    self.cache = resource_template
        else:
            self.cache = resource_template

    def save(self):
        with codecs.open(self.cache_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.cache, ensure_ascii=False, indent=4))
            f.close()

    def download(self, uri):
        self.cache['uri'] = uri
        self.save()

        if self.download_dir.exists():
            shutil.rmtree(self.download_dir.absolute())
        self.download_dir.mkdir(exist_ok=True)

        uri = urlparse(self.cache.get('uri'))
        if uri.scheme == 'http' or uri.scheme == 'https':
            self._http()
        elif uri.scheme.startswith('git+'):
            self._git()
        else:
            raise RescourceException(f'Unknown scheme {self.cache}')
        
    def _git(self):
        git_url = self.cache.get('uri')[4:]
        p = subprocess.run(f'git clone {git_url} {self.download_dir.absolute()}', shell=True)
        p.check_returncode()
        logger.warning(f'Source downloaded to {str(self.download_dir.absolute())}')
    
    def _http(self):
        # resp = requests.get(self.cache.get('uri'), allow_redirects=True)
        # TODO support http download 
        # 1. download gzip file and unzip it
        # 2. download from git repo
        pass


class RescourceException(Exception):
    pass



import codecs
import copy
import json
import os
import shutil
from functools import reduce
from urllib import parse as url_parser
from typing import Dict
from pathlib import Path
import traceback
from flask import Response
from lyrebird import log
from . import context
import sys


_logger = log.get_logger()

"""
文件管理系统

用于管理存储在文件中mock数据，数据组

"""

class FileSystemError(Exception):
    pass


class FileManager:
    """
    FileSystem入口
    
    管理整个mock服务器的数据文件
    
    数据文件以数据组的形式保存于数据目录下，每个数据组下包含若干组数据和一个匹配规则文件。
    数据组之间可以继承[注意:不要循环继承]。
    """
    def __init__(self):
        self.data_dir = None
        self.data_groups: Dict[str, DataGroup] = dict()
        self.current_data_group = None

    def set_temp_group(self, path):
        temp_group = DataGroup.create_from_dir(path)
        self.current_data_group = temp_group

    def set_root(self, path):
        data_root = Path(path).absolute()
        if not data_root.exists():
            try:
                data_root.mkdir(parents=True)
            except Exception:
                _logger.error(f'Data root path not found. And make new dir failed. {path}')
                raise FileSystemError(f'Data root path not found. And make new dir failed. {path}')
        if not os.path.isdir(path):
            raise FileSystemError(f'Set data root path fail. {path} is not a dir')
        self.data_dir = path
        self.scan()

    def scan(self):
        self.data_groups.clear()
        # 由于数据组只存在于数据目录下，因此没有递归查找的处理
        data_group_dir_names = os.listdir(self.data_dir)
        for dir_name in data_group_dir_names:
            group = DataGroup.create_from_dir(os.path.join(self.data_dir, dir_name))
            if group:
                self.data_groups[group.name] = group

    def add_group(self, name, json_obj):
        """
        新建数据组

        如果数据组已存在，则新数据与旧数据组合并

        """
        group_abspath = os.path.abspath(os.path.join(self.data_dir, name))
        if not os.path.exists(group_abspath):
            os.mkdir(group_abspath)
        new_group = DataGroup.create_from_data(group_abspath, json_obj)
        old_group = self.data_groups.get(name)
        if old_group:
            old_group.merge(new_group)
        else:
            self.data_groups[new_group.name] = new_group
        self.scan()

    def update_group(self, origin_name, name, json_obj):
        """
        更新数据组

        """
        group = self.data_groups.get(origin_name)
        if group:
            group.conf = json_obj
            group.write_conf()
            if origin_name != name:
                group.rename(name)
        self.scan()

    def set_current_data_group(self, name):
        if not name or name == '':
            self.current_data_group = None
            return True
        target_group = self.data_groups.get(name)
        if target_group:
            self.current_data_group = target_group
            return True
        else:
            return False


class DataGroup:
    """
    Mock数据组
    
    根据数据组目录下conf.json文件,判断请求是否匹配某一条mock数据。
    如果匹配则返回mock数据中的response。
    如果不匹配则返回None。
    
    """
    CONF = 'conf.json'

    @classmethod
    def create_from_data(cls, dir_path, json_data):
        group = cls(dir_path)
        group.conf = json_data
        group.write_conf()
        group.scan()
        return group

    @classmethod
    def create_from_dir(cls, dir_path):
        if not os.path.isdir(dir_path):
            return None
        sub_file_names = os.listdir(dir_path)
        if DataGroup.CONF not in sub_file_names:
            return None
        group = cls(dir_path)
        group.read_conf()
        group.scan()
        return group

    def __init__(self, dir_path):
        self.dir_path = dir_path
        self.name = os.path.basename(dir_path)
        self.conf_path = os.path.abspath(os.path.join(dir_path, DataGroup.CONF))
        self.conf = None
        self.data_dict: Dict[str, Data] = {}

    def scan(self):
        self.data_dict = {}
        sub_file_names = os.listdir(self.dir_path)
        for file_name in sub_file_names:
            data = Data.create_from_dir(os.path.join(self.dir_path, file_name))
            if data:
                self.data_dict[data.name] = data

    def merge(self, group):
        self.conf['filters'] += group.conf['filters']
        self.conf['parent'] = group.conf['parent']
        self.write_conf()

    def rename(self, name):
        new_path = os.path.join(os.path.dirname(self.dir_path), name)
        os.rename(self.dir_path, new_path)
        self.name = name
        self.dir_path = os.path.abspath(new_path)
        self.conf_path = os.path.abspath(os.path.join(new_path, DataGroup.CONF))

    def delete(self):
        shutil.rmtree(self.dir_path)

    def read_conf(self):
        try:
            self.conf = json.loads(codecs.open(self.conf_path, 'r', 'utf-8').read())
        except Exception as e:
            exc_info = sys.exc_info()
            _logger.error(f'Read conf error. \nFile:{self.conf_path}. \nClass:{exc_info[0]} \nMessage:{exc_info[1]}')
            raise FileSystemError(f'Read conf error. File={self.conf_path}') from e

    def write_conf(self):
        f = codecs.open(self.conf_path, 'w', 'utf-8')
        conf_str = json.dumps(self.conf, ensure_ascii=False, indent=4)
        f.write(conf_str)
        f.close()

    def get_response(self, url, requset_data = None):
        for req_filter in self.conf['filters']:
            contents = req_filter['contents']
            # 跳过空的过滤条件
            if len(contents) == 0:
                continue
            # 检查filters/contents中的字符是否都在url中
            hit = True
            for content in contents:
                if content not in url:
                    hit = False
                    break
            # 判断filters／request是否存在（手工配置），且当前url的requset_data不为空，则进行对比校验
            if hit and 'body_filter' in req_filter.keys() and requset_data != None:
                try:
                    if req_filter['body_filter'] not in requset_data.decode():
                        hit = False
                except Exception:
                    traceback.print_exc()
                    _logger.error(f'请检查config文件，URL为{contents}是否配置了body_filter字段，会触发对请求参数的校验!')
            # 如果未命中继续查找
            if not hit:
                continue
            # 命中, 返回mock数据
            data = self.data_dict[req_filter['response']]
            resp_headers = [('lyrebird', 'mock;'+data.name)]
            for item in data.resp_headers:
                if isinstance(item, str):
                    name = item
                    value = data.resp_headers[name]
                elif isinstance(item, list):
                    name = item[0]
                    value = item[1]
                else:
                    continue
                # rm 'content-length' from ignore list
                if name.lower() in ('connection',
                                    'content-encoding',
                                    'transfer-encoding'):
                    continue
                resp_headers.append((name, value))
            return Response(data.resp_data, data.resp_code, resp_headers)
        # 如果conf中配置了parent，开始递归查找
        if self.conf.get('parent'):
            parent_path = self.conf['parent']
            if not os.path.isabs(parent_path):
                parent_path = os.path.abspath(os.path.join(self.dir_path, parent_path))
            if not os.path.exists(parent_path):
                raise FileSystemError('Can not find parent data group')
            parent_group = DataGroup.create_from_dir(parent_path)
            return parent_group.get_response(url)
        # 没有匹配任何数据
        return None

    def add_data(self, name, json_data):
        data_dir = os.path.join(self.dir_path, name)
        if not os.path.exists(data_dir):
            os.mkdir(data_dir)
        Data.create_from_data(data_dir, json_data)
        self.scan()

    def update_data(self, origin_name, name, json_data):
        data = self.data_dict.get(origin_name)
        data.json_data = json_data
        data.write_file()
        data.rename(name)
        self.scan()
    
    def add_data_and_filter(self, flow):
        url = flow['request']['url']
        result = url_parser.urlparse(url)
        resp_name = result.path.replace('/', '.')[1:]+'_'+flow['id']
        self.add_data(resp_name, flow)
        # TODO make simple filters.
        if result.path == '' or result.path == '/':
            filter_contents = [result.hostname]
        else:
            filter_contents = [result.path]
        self.conf['filters'].append({'contents':filter_contents, 'response':resp_name})
        self.write_conf()
        self.scan()

class Data:
    """
    Mock数据文件管理
    
    request.json
    >保存请求头，URL
    
    request_data.json
    request_data.form
    request_data.bin
    >保存请求体，根据不同格式扩展名不同
    
    response.json
    >响应状态码，响应头
    
    response_data.json
    response_data.bin
    >响应数据，根据不同格式扩展名不同
    
    """
    @classmethod
    def create_from_data(cls, data_dir, json_data):
        data = cls(data_dir)
        data.json_data = json_data
        data.write_file()
        return data

    @classmethod
    def create_from_dir(cls, data_dir):
        if not os.path.isdir(data_dir):
            return None
        return cls(data_dir)

    def __init__(self, path):
        self.name = os.path.basename(path)
        self.path = path
        self.json_data = None

    @property
    def resp_data(self):
        contents = os.listdir(self.path)
        if 'response_data.json' in contents:
            resp_body_file = os.path.join(self.path, 'response_data.json')
        elif 'response_data.bin' in contents:
            resp_body_file = os.path.join(self.path, 'response_data.bin')
        else:
            return ''
        f = codecs.open(resp_body_file, 'r', 'utf-8')
        res = f.read()
        f.close()
        return res

    @property
    def resp_headers(self):
        f = codecs.open(os.path.join(self.path, 'response.json'))
        resp = json.loads(f.read())
        f.close()
        return resp['headers']

    @property
    def resp_code(self):
        f = codecs.open(os.path.join(self.path, 'response.json'))
        resp = json.loads(f.read())
        f.close()
        return resp['code']

    def rename(self, name):
        new_path = os.path.join(os.path.dirname(self.path), name)
        os.rename(self.path, new_path)
        self.name = name
        self.path = new_path

    def write_file(self):
        if not self.json_data:
            raise FileSystemError('Write to file error. Not set json data')
        req = self.json_data.get('request')
        if req:
            req_header = {'url': req.get('url'), 'method': req.get('method', 'GET'), 'headers': req.get('headers')}
            if req_header:
                self._write_file('request', req_header)
            req_data = req.get('data')
            if req_data:
                self._write_file('request_data', req_data)
        resp = self.json_data.get('response')
        if resp:
            # 保存时去掉数据源标记
            resp_header = {'code': resp.get('code', 200), 'headers': copy.deepcopy(resp.get('headers'))}
            resp_header['headers'].pop('lyrebird', None)
            if resp_header:
                self._write_file('response', resp_header)
            resp_data = resp.get('data')
            if resp_data:
                self._write_file('response_data', resp_data)

    def _write_file(self, name, data):
        if isinstance(data, (str, bytes)):
            name = name + '.bin'
            if isinstance(data, str):
                data = data.encode()
            f = codecs.open(os.path.join(self.path, name), 'wb')
            f.write(data)
            f.close()
        elif isinstance(data, (dict, list)):
            name = name + '.json'
            data = json.dumps(data, ensure_ascii=False, indent=4)
            f = codecs.open(os.path.join(self.path, name), 'w', 'utf-8')
            f.write(data)
            f.close()
        else:
            raise FileSystemError(f'Write to file error. Unsupported type {type(data)}')

    def read_file(self):
        self.json_data = dict()
        req_h = self._read_file('request')
        req_data = self._read_file('request_data')
        self.json_data['request'] = req = dict()
        if req_h:
            req['url'] = req_h.get('url')
            req['method'] = req_h.get('method')
            req['headers'] = req_h.get('headers')
        if req_data:
            req['data'] = req_data
        resp_h = self._read_file('response')
        resp_data = self._read_file('response_data')
        self.json_data['response'] = resp = dict()
        if resp_h:
            resp['code'] = resp_h.get('code')
            resp['headers'] = resp_h.get('headers')
        if resp_data:
            resp['data'] = resp_data

    def _read_file(self, name):
        file_path = os.path.join(self.path, name+'.json')
        if os.path.exists(file_path):
            return json.loads(codecs.open(file_path, 'r', 'utf-8').read())
        file_path = os.path.join(self.path, name+'.bin')
        if os.path.exists(file_path):
            return codecs.open(file_path, 'rb').read().decode()

    def delete(self):
        shutil.rmtree(self.path)


class DataGroupConfBuilder:

    def __init__(self):
        self.host = False
        self.path = True
        self.split_path = False
        self.query = False
        self.split_query = False
        self._conf = dict()
        self._conf['parent'] = None
        self._conf['filters'] = []

    def add_filter_by_req_ctx(self, req_ctx_dict):
        conf = context.application.conf
        _filter = dict()
        url_filter_contents = []
        ctx_id = req_ctx_dict['id']
        url = req_ctx_dict['request']['url']
        result = url_parser.urlparse(url)
        ''' hotfix new conf did not contains record settings
        if conf.get('mock') and conf.get('mock').get('record'):
            record = conf.get('mock').get('record')
            self.host = record.get('host', self.host)
            self.path = record.get('path', self.path)
            self.split_path = record.get('split_path', self.split_path)
            self.query = record.get('query', self.query)
            self.split_query = record.get('split_query', self.split_query)
        '''
        if self.host:
            url_filter_contents.append(result.hostname)
        if self.path and len(result.path) > 1:
            if self.split_path:
                path_items = result.path.split('/')
                url_filter_contents += path_items
            else:
                url_filter_contents.append(result.path)
        if self.query and len(result.query) > 0:
            if self.split_query:
                query_items = result.query.split('&')
                url_filter_contents += query_items
            else:
                url_filter_contents.append(result.query)
        _filter['contents'] = list(filter(lambda x: True if len(x) > 0 else False, url_filter_contents))
        _filter['response'] = result.path.replace('/', '.')[1:] + '_' + ctx_id
        self._conf['filters'].append(_filter)
        return self

    def build(self):
        return self._conf


class DataHelper:

    @staticmethod
    def to_dict(req, req_data, resp, resp_data):
        data = dict()
        req_dict = data['request'] = dict()
        if req:
            req_dict.update(json.loads(req))
        if req_data:
            try:
                req_data = json.loads(req_data)
            except Exception:
                pass
            req_dict['data'] = req_data
        resp_dict = data['response'] = dict()
        if resp:
            resp_dict.update(json.loads(resp))
        if resp_data:
            try:
                resp_data = json.loads(resp_data)
            except Exception:
                pass
            resp_dict['data'] = resp_data
        return data



from lyrebird import application
from lyrebird.mock import context

class NoticeCenter():
    
    def __init__(self):
        # display notice in frontend
        application.server['event'].subscribe('notice', self.push_notice)

    def push_notice(self, msg):
        """
        display notice
        
        """
        context.application.socket_io.emit('show', msg)



from flask import Blueprint
from flask_restful import Resource, Api
from .common import Status, WorkMode, Manifest
from .flow import Flow, FlowList
from .mock import MockData, MockGroup, ActivatedMockGroup , MockGroupByName
from .config import Conf, ResetConf
from .plugin import Plugin
from .menu import Menu
from .event import Event


api = Blueprint('api', __name__, url_prefix='/api')
api_source = Api(api)


api_source.add_resource(Status, '/status')
api_source.add_resource(Manifest, '/manifest')
api_source.add_resource(Flow, '/flow/<string:id>')
api_source.add_resource(FlowList, '/flow')
api_source.add_resource(MockGroup, '/mock', '/mock/<string:group_id>')
api_source.add_resource(MockData, '/mock/<string:group_id>/data', '/mock/<string:group_id>/data/<string:data_id>')
api_source.add_resource(ActivatedMockGroup, '/mock/activated', '/mock/<string:group_id>/<string:action>')
api_source.add_resource(MockGroupByName, '/mock_by_name')
api_source.add_resource(Conf, '/conf/<string:plugin_name>')
api_source.add_resource(ResetConf, '/conf/<string:plugin_name>/reset')
api_source.add_resource(WorkMode, '/mode', '/mode/<string:mode>')
api_source.add_resource(Plugin, '/plugin', '/plugin/<string:plugin_name>')
api_source.add_resource(Menu, '/menu')
api_source.add_resource(Event, '/event')



from flask import render_template, Blueprint
from bs4 import BeautifulSoup
from ..reporter import report_handler
from .. import plugin_manager


plugin = Blueprint('plugin', __name__, url_prefix='/plugin', template_folder='../templates', static_folder='../static')


@plugin.route('/<string:name>')
def plugin_base(name):
    report_handler.page_in(name)
    plugin = plugin_manager.plugins.get(name)
    if not plugin:
        return "Plugin not found"
    ui = plugin['beta_web']
    web_content = ui.index()
    soup = BeautifulSoup(web_content, 'html.parser')

    # set all javascripts into script block
    all_scripts = []
    for javascript_tag in soup.find_all('script'):
        all_scripts.append(javascript_tag)
        javascript_tag.extract()
    # set all css link into header
    all_css = []
    for css_tag in soup.find_all('link'):
        all_css.append(css_tag)
        css_tag.extract()
    return render_template('plugin.html', 
                            plugin_content=str(soup),
                            plugin_javascript=all_scripts,
                            plugin_css=all_css)




"""
Base threading server class
"""

from threading import Thread


class ThreadServer:

    def __init__(self):
        self.server_thread = None
        self.running = False

    def start(self, *args, **kwargs):
        if self.running:
            return
        self.running = True
        self.server_thread = Thread(target=self.run, args=args, kwargs=kwargs)
        self.server_thread.start()

    def stop(self):
        self.running = False

    def run(self):
        """
        Server main function
        """
        pass



import argparse
import webbrowser
import json
import traceback
import socket
import threading
import signal
import os
from pathlib import Path
from lyrebird import log
from lyrebird import application
from lyrebird.config import Rescource, ConfigManager
from lyrebird.mock.mock_server import LyrebirdMockServer
from lyrebird.proxy.proxy_server import LyrebirdProxyServer
from lyrebird.event import EventServer
from lyrebird.task import BackgroundTaskServer
from lyrebird.notice_center import NoticeCenter
from lyrebird.db.database_server import LyrebirdDatabaseServer


logger = log.get_logger()


def main():
    """
    Command line main entry

    Start lyrebird

    * start in default config
    ```
    lyrebird
    ```
    * start with verbose mode
    ```
    lyrebird -v
    ```
    * start without open a web browser
    ```
    lyrebird -b
    ```
    * start with a specified config file
    ```
    lyrebird -c /path/to/your/config/file
    ```
    * start with multipart args
    ```
    lyrebird -v --mock 8080 -c /path/to/your/config/file
    ```
    """
    parser = argparse.ArgumentParser(prog='lyrebird')

    parser.add_argument('-v', dest='verbose', action='store_true', help='Show verbose log')
    parser.add_argument('--mock', dest='mock', type=int, help='Set mock server port, default port is 4272')
    parser.add_argument('--proxy', dest='proxy', type=int, help='Set proxy server port, default port is 9090')
    parser.add_argument('--data', dest='data', help='Set data dir, default is "./data/"')
    parser.add_argument('-b', '--no_browser', dest='no_browser', action='store_true', help='Start without open a browser')
    parser.add_argument('-c', '--config', dest='config', help='Start with a config file. Default is "~/.lyrebird/conf.json"')
    parser.add_argument('--log', dest='log', help='Set output log file path')

    subparser = parser.add_subparsers(dest='sub_command')
    src_parser = subparser.add_parser('src')
    src_parser.add_argument('uri')
    subparser.add_parser('plugin')

    args = parser.parse_args()

    if args.config:
        application._cm = ConfigManager(conf_path=args.config)
    else:
        application._cm = ConfigManager()
    application._src = Rescource()

    # set current ip to config
    try:
        application._cm.config['ip'] = _get_ip()
    except socket.gaierror as e:
        logger.error('Failed to get local IP address, error occurs on %s' % e)

    if args.verbose:
        application._cm.config['verbose'] = True

    # init file logger after config init
    log.init(args.log)
    
    if args.mock:
        application._cm.config['mock.port'] = args.mock
    if args.proxy:
        application._cm.config['proxy.port'] = args.proxy
    if args.data:
        application._cm.config['mock.data'] = args.data

    logger.debug(f'Read args: {args}')

    if args.sub_command == 'src':
        logger.debug('EXEC SUBCMD:SRC')
        src(args)
    elif args.sub_command == 'plugin':
        logger.debug('EXEC SUBCMD:PLUGIN')
        plugin(args)
    else:
        logger.debug('EXEC LYREBIRD START')
        run(args)


def run(args:argparse.Namespace):
    # Check mock data group version. Update if is older than 1.x
    from . import mock_data_formater
    data_path = application._cm.config['mock.data']
    data_dir = Path(data_path)
    mock_data_formater.check_data_dir(data_dir)

    # show current config contents
    config_str = json.dumps(application._cm.config, ensure_ascii=False, indent=4)
    logger.warning(f'Lyrebird start with config:\n{config_str}')

    application.server['event'] = EventServer()
    application.server['task'] = BackgroundTaskServer()
    application.server['proxy'] = LyrebirdProxyServer()   
    application.server['mock'] = LyrebirdMockServer()
    application.server['db'] = LyrebirdDatabaseServer()

    application.start_server()

    # activate notice center
    application.notice = NoticeCenter()
    
    # auto open web browser
    if not args.no_browser:
        webbrowser.open(f'http://localhost:{application.config["mock.port"]}')

    # stop event handler
    def signal_handler(signum, frame):
        application.stop_server()
        threading.Event().set()
        logger.warning('!!!Ctrl-C pressed. Lyrebird stop!!!')
        os._exit(1)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)


def debug():
    # use lyrebird.debug to start plugin in debug mode
    # can pass args by sys.args
    main()
    # main thread loop
    import asyncio
    loop = asyncio.get_event_loop()
    loop.run_forever()


def plugin(args:argparse.Namespace):
    pass


def src(args:argparse.Namespace):
    from threading import Thread
    def worker():
        application._src.download(args.uri)
    Thread(target=worker).start()


def _get_ip():
    """
    Get local ip from socket connection

    :return: IP Addr string
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('bing.com', 80))
    return s.getsockname()[0]



from setuptools import setup


setup(
    name='lyrebird-demo',
    version='0.1.0',
    packages=['lyrebird_demo'],
    url='',
    license='',
    author='hbqa',
    author_email='',
    description='',
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
        ],
        'lyrebird_data_handler': [
            'demo = lyrebird_demo.my_handler:MyDataHandler'
        ],
        'lyrebird_web': [
            'demo = lyrebird_demo.my_ui:MyUI'
        ]
    },
    install_requires=[
        'requests'
    ]

)
